/*
This is a limited iff-ilbm loader
Copyright (C) 2000  Tim Doty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 drakki@rollanet.org

Code is adapted from pcx.cc found in gocr-0.2.7 by Joerg Schulenburg
and was written for use with gocr.

 */

#include <stdio.h>
#include <stdlib.h>

#include "iff.h"

typedef unsigned char byte;

#define ERR(x) { fprintf(stderr,"ERROR "__FILE__" L%d: " x "\n",__LINE__);exit(1);}

int errn;
// used to read file aware of eof
char read_byte(FILE *f1){
  char c=0; c=fgetc(f1);
  if(feof(f1))errn=1;
  if(ferror(f1))errn+=2; return c;
}

void Quit(int iLoc, int x, int y) {
if(errn%2==1) {
  fprintf(stderr,"\nEOF(%d) at x=%d y=%d\n",iLoc,x,y);
  }
if(errn>1) {
  perror("\nfile error(2): ");
  }
if(errn==0) {
  fprintf(stderr,"\nsign error(%d): x=%d, y=%d\n",iLoc,x,y);
  }
exit(5);
}

/* only reads b&w ilbms */
void readilbm(char *name,pix *p, int vvv) {
    int page,nx,ny,i,j,x,y,bits,pal[256][3];
    FILE *f1;  
    unsigned char *pic,h[128],bb,b1;
    errn=0;
    BitMapHeader *BMHD;
    errn=0;
    for(i=0;i<256;i++)for(j=0;j<3;j++)pal[i][j]=i; /* inits grey scale */
    f1=fopen(name,"rb"); if(!f1) ERR("open");
    if(fread(h,1,20,f1)!=20)ERR("read IFF-ILBM header"); /* FORM????ILBMBMHD */
    if(h[0]!='F'||h[1]!='O'||h[2]!='R'||h[3]!='M')ERR("not IFF");
    if(h[8]!='I'||h[9]!='L'||h[10]!='B'||h[11]!='M')ERR("not ILBM");
    if(h[12]!='B'||h[13]!='M'||h[14]!='H'||h[15]!='D')ERR("Bit Map Header not found");
    fread(BMHD,20,1,f1);
    bits = BMHD->nPlanes;
    ny = BMHD->h;
    nx = BMHD->w;
    if(bits!=1 && bits!=8)ERR("Unsupported number of bitplanes");
    if(BMHD->compression>1)ERR("Unsupported compression");

    /* Doesn't currently extract DPI chunks. Who cares? */
    if(vvv) /* are we verbose? */
        fprintf(stderr,"# IFF-ILBM bits=%d x=%d y=%d",bits,nx,ny);
    fflush(stdout);

    while(read_byte(f1)!='B' && errn==0);
    if(errn>0) {
        ERR("BODY chunk not found");
        }
    if(read_byte(f1)!='O') errn=1;
    if(read_byte(f1)!='D') errn=1;
    if(read_byte(f1)!='Y') errn=1;
    if(errn>0) {
        ERR("BODY chunk not found");
        Quit(5,-1,-1);
        }
    pic=new unsigned char[ nx*ny ];
    if(pic==NULL)ERR("Out of memory");
    /* initialize memory space */
    for(x=0;x<nx*ny;x++)
        pic[x] = 0;
    x=y=0;
    /* bypass body chunk size */
    read_byte(f1);read_byte(f1);read_byte(f1);read_byte(f1);
    if(BMHD->compression==0) {
        for(y=0;y<ny;y++) { /* do following for each row */
            for(page=0;page<bits;page++) { /* page means plane */
                x = 0; /* set x for beginning of each row in each plane */
                while(x<nx) {
                    bb = read_byte(f1);
                    if(errn>0) Quit(0,x,y);
                    if(x+7<nx) {
                        pic[7+x+nx*y] += (unsigned char)((bb&1)<<page);
                        bb = bb>>1;
                        pic[6+x+nx*y] += (unsigned char)((bb&1)<<page);
                        bb = bb>>1;
                        pic[5+x+nx*y] += (unsigned char)((bb&1)<<page);
                        bb = bb>>1;
                        pic[4+x+nx*y] += (unsigned char)((bb&1)<<page);
                        bb = bb>>1;
                        pic[3+x+nx*y] += (unsigned char)((bb&1)<<page);
                        bb = bb>>1;
                        pic[2+x+nx*y] += (unsigned char)((bb&1)<<page);
                        bb = bb>>1;
                        pic[1+x+nx*y] += (unsigned char)((bb&1)<<page);
                        bb = bb>>1;
                        pic[x+nx*y]   += (unsigned char)((bb&1)<<page);
                        x+=8;
                        }
                    else {
                        for(int k=7;k>=0;k--) {
                            if(k<nx-x)
                                pic[k+x+nx*y] += (unsigned char)((bb&1)<<page);
                            bb = bb>>1;
                            }
                        x = nx;
                        }
                    }
                }
            }
        }
    else { /* uses runlength encoding */
        for(y=0;y<ny;y++) { /* do following for each row */
            for(page=0;page<bits;page++) { /* page means plane */
                x = 0; /* set x for beginning of each row in each plane */
                while(x<nx) {
                    bb = read_byte(f1);
                    if(errn>0) Quit(1,x,y);
                    j = (bb>127) ? bb-256 : bb;
                    if(j>=0) {  /* read j+1 bytes literally */
                        for(i=0;i<=j;i++) {
                            bb = read_byte(f1);
                            if(errn>0) Quit(2,x,y);
                            if(x+7<nx) {
                                pic[7+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[6+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[5+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[4+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[3+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[2+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[1+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[x+nx*y]   += (unsigned char)((bb&1)<<page);
                                x+=8;
                                }
                            else {
                                for(int k=7;k>=0;k--) {
                                    if(k<nx-x)
                                        pic[k+x+nx*y] += (unsigned char)((bb&1)<<page);
                                    bb = bb>>1;
                                    }
                                x = nx;
                                }
                            }
                        }
                    else if(j>-128) { /* replicate byte -j+1 times */
                        b1 = read_byte(f1);
                        if(errn>0) Quit(3,x,y);
                        for(i=0;i>=j;i--) {
                            bb = b1;
                            if(x+7<nx) {
                                pic[7+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[6+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[5+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[4+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[3+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[2+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[1+x+nx*y] += (unsigned char)((bb&1)<<page);
                                bb = bb>>1;
                                pic[x+nx*y]   += (unsigned char)((bb&1)<<page);
                                x+=8;
                                }
                            else {
                                for(int k=7;k>=0;k--) {
                                    if(k<nx-x)
                                        pic[k+x+nx*y] += (unsigned char)((bb&1)<<page);
                                    bb = bb>>1;
                                    }
                                x = nx;
                                }
                            }
                        }
                    }
                }
            }
        }
    /* convert 1-bit to 8-bit */
    if(bits==1) {
        for(x=0;x<nx*ny;x++) {
            pic[x] = (pic[x]==1) ? 255 : 0;
            }
        }
    fclose(f1);
    p->p=pic;  p->x=nx;  p->y=ny; p->bpp=1;
    if(vvv)fprintf(stderr,"\n");
    }
