/*
This is a limited iff-ilbm loader
Copyright (C) 2000  Tim Doty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 drakki@rollanet.org

Code is adapted from pcx.cc found in gocr-0.2.7 by Joerg Schulenburg
and was written for use with gocr.

 */

#ifndef __IFF_H_
#define __IFF_H_

#include "pnm.h"

typedef unsigned char UBYTE;    /*  8 bits unsigned     */
typedef short WORD;             /* 16 bits signed       */
typedef unsigned short UWORD;   /* 16 bits unsigned     */
typedef long LONG;              /* 32 bits signed       */

typedef LONG ID;                /* 4 chars in ' ' through '~'   */
#define MakeID(a,b,c,d) ( (a)<<24 | (b)<<16 | (c)<<8 | (d) )

typedef struct {
        LONG    ckID;
        LONG    ckSize;         /* sizeof(ckData)       */
        UBYTE  *ckData;         /* ckSize */
        } Chunk;

typedef UBYTE Masking;  /* Choice of masking technique. */

#define mskNone 0
#define mskHasMask      1
#define mskHasTransparentColor  2
#define mskLasso        3

typedef UBYTE Compression;      
        /* Choice of compression algorithm applied to the rows of all 
         * source and mask planes. "cmpByteRun1" is the byte run encoding 
         * described in Appendix C. Do not compress across rows! */
#define cmpNone 0
#define cmpByteRun1     1

typedef struct {
        UWORD w, h;                   /* raster width & height in pixels      */
        WORD  x, y;                   /* pixel position for this image        */
        UBYTE nPlanes;                /* # source bitplanes                   */
        Masking masking;
        Compression compression;
        UBYTE pad1;                   /* unused; for consistency, put 0 here  */
        UWORD transparentColor;       /* transparent "color number" (sort of) */
        UBYTE xAspect, yAspect;       /* pixel aspect, a ratio width : height */
        WORD  pageWidth, pageHeight;  /* source "page" size in pixels         */
        } BitMapHeader;

void readilbm(char *name, pix *p, int vvv);
Chunk GetChunk(FILE *f1);

#endif
