/*
This is a Optical-Character-Recognition program
Copyright (C) 2000  Joerg Schulenburg

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 Joerg.Schulenburg@physik.uni-magdeburg.de

  sometimes I have written comments in german language, sorry for that

 This file was retrieved from pgm2asc.cc of Joerg, in order to have
 a library of the ocr-engine from Klaas Freitag <freitag@suse.de>
 
 Thats todo from the libs ;)        
 ToDo: - add filter (r/s mismatch) g300c1
       - write parallelizable code!
       - learnmode (optimize filter)
       - use ispell for final control or if unsure
       - better line scanning (if not even)
       - step 5: same chars differ? => expert mode
       - chars dx>dy and above 50% hor-crossing > 4 is char-group ?
       - detect color of chars and underground
       - rotation of chars in the pixmap to avoid y-corrections
  

 */
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "pgm2asc.h"
#include "pcx.h"
#include "ocr0.h"	// only_numbers


void help( void );
void out_version(int v);

// -------------------------------------------------------------
// ------   MAIN - replace this by your own aplication! 
// -------------------------------------------------------------
int main(int argn, char *argv[]){
  int i,cs=0,spc=0,mo=0,dust_size=10,numo=0;
  int verbose=0;
  int linecounter = 0;
  const char *line;
  
  pix p;  char *inam="-",cc=0,*lc="_",*s1;
  setvbuf(stdout,(char *)NULL,_IONBF,0); // not buffered
  if(argn<=1){ out_version(1); exit(0); }
  for(i=1;i<argn;i++){
    if(strcmp(argv[i],"--help") == 0){ help();continue; }
    if( argv[i][0]=='-' && argv[i][1]!=0 ) { cc=argv[i][1];
      s1="";if(i+1<argn)s1=argv[i+1];
      switch ( cc ) {
       case 'h' : help(); break;
       case 'i' : inam =  s1;i++; break;
       case 'e' : if(s1[0]=='-' && s1[1]==0){
                    (*stderr)=(*stdout); /* are there better solutions ??? */
                  } else
                  if(!freopen(s1,"w",stderr)){
                    fprintf(stderr,"stderr redirection to %s failed\n",s1);
                  };i++; break;
       case 'o' : if(s1[0]=='-' && s1[1]==0){ /* default */ } else
                  if(!freopen(s1,"w",stdout)){
                    fprintf(stderr,"stdout redirection to %s failed\n",s1);
                  };i++; break;
       case 'c' : lc   =  s1;i++; break;
       case 'd' : dust_size  =  atoi(s1);i++; break;
       case 'l' : cs  =  atoi(s1);i++; break;
       case 's' : spc =  atoi(s1);i++; break;
       case 'v' : verbose |= atoi(s1);i++; break;
       case 'm' : mo  |= atoi(s1);i++; break;
       case 'n' : numo=atoi(s1);i++; break;
       default: fprintf(stderr,"# unknown option use -h for help\n");
      } cc=0; continue;
    } else /* argument can be filename v0.2.5 */
    if( argv[i][0]!='-' || argv[i][1]==0 ) { inam=argv[i]; }
  }
  if(!cs) cs =128+32;
  /* if(verbose)setvbuf(stdout,NULL,_IONBF,0); */
  if(verbose){
    fprintf(stderr,"#");out_version(0);
    fprintf(stderr,"# options are: -l %d -s %d -v %d -c %s -m %d -d %d -n %d %s\n",
             cs,spc,verbose,lc,mo,dust_size,numo,inam);
  }
  set_only_numbers(numo);
  // ----- read picture
  if(strstr(inam,".pcx")) { readpcx(inam,&p,verbose); }
  else                    { readpgm(inam,&p,verbose); }

  // separation of main and rest for using as lib
  // this will be changed later => introduction of set_option()
  //   for better communication to the engine 

  pgm2asc(&p, mo, cs, spc, dust_size, lc, verbose);

  linecounter = 0;
  line = getTextLine( linecounter++ );
  while( line ) {
     printf( "%s\n", line );
     line = getTextLine( linecounter++ );
  }
  free_textlines();
  
  delete p.p;
}


void help( void ){
  out_version(0);
  fprintf(stderr,
   " using: gocr [options] pnm_file_name  # use - for stdin\n"
   " options:\n"
   " -h        - get this help\n"
   " -i name   - input image file (pnm,pgm,pbm,ppm,pcx,...)\n"
   " -i -      - read PNM from stdin (djpeg -gray a.jpg | gocr -)\n"
   " -o name   - output file  (redirection of stdout)\n"
   " -e name   - logging file (redirection of stderr)\n"
   " -l num    - grey level 0<160<=255\n"
   " -d num    - dust_size (remove all smaller clusters, 10=default)\n"
   " -s num    - spacewidth/dots (0 = autodetect)\n"
   " -v num    - verbose  [summed]\n"
   "      1      print more info\n"
   "      2      list chapes  of boxes (see -c)\n"
   "      4      list pattern of boxes (see -c)\n"
   "      8      print pattern after recognition\n"
   "     16      print line infos\n"
   "     32      debug outXX.pgm\n"
   " -c string - list of chars (_ = not recognized chars)\n"
   " -m num    - operation modes, ~ = switch off\n"
   "      2      use database (early development)\n"
   "      4      layout analysis, zoning (development)\n"
   "      8      ~ compare non recognized chars\n"
   "     16      ~ divide overlapping chars\n"
   "     32      ~ context correction\n"
   "     64      char packing (development)\n"
   " -n   1      only numbers\n"
   " examples:\n"
   "\tgocr -v 33 text1.pbm                # some infos + out30.bmp\n"
   "\tgocr -v 7 -c _YV text1.pbm          # list unknown, Y and V chars\n"
   "\tdjpeg -pnm -gray text.jpg | gocr -  # use jpeg-file via pipe\n"
   "\n");
  exit(0);
}

#define Version "v0.2.7  2000/08/01"

void out_version(int v){
  fprintf(stderr," Optical Character Recognition --- gocr "Version"\n");
  if(v)  fprintf(stderr," use option -h for help\n");
  if(v&2) exit(1);
  return;
}
